/*
 *  Copyright (C) 2006 David Kozub <zub at linux.fjfi.cvut.cz.nospam>
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

/*
  A simple wrapper around stdio stream. Keeps track of current line number,
  can skip #-style comments etc...
*/

#include <stdio.h>
#include <stdlib.h>
#include "uni_str.h"
#include "trans_tbl.h"
#include "tbl_parse.h"

class file {
  FILE *f;
  int line;

public:
  file(FILE *f);
  ~file();

  int getc();
  int ungetc(int c);
  inline int skip_ws_fgetc();

  inline bool check_next_char_skip_ws(char c);
  inline bool check_next_char(char c);
  bool check_str_skip_ws(const char *s);

  bool parse_int(int &i);
  bool parse_uni(unsigned &i);
  bool parse_uni_str(uni_str &s);

  int get_line_no();

  bool ferror();
  bool feof();
};

file::file(FILE *f): line(1)
{
  this->f=f;
}

file::~file()
{
  if (f) fclose(f);
}

int file::getc()
{
  char c=fgetc(f);
  if (c=='\n') line++;

  // skip comments
  if (c=='#') {
    do {
      c=fgetc(f);
    } while (c!=EOF && c!='\n');
    if (c=='\n') line++;
  }
  
  return c;
}

int file::ungetc(int c)
{
  if (c=='\n') line--;

  return ::ungetc(c,f);
}

int file::skip_ws_fgetc()
{
  int c=getc();
  while (c!=EOF && c<=' ') c=getc();
  
  return c;
}

bool file::check_next_char_skip_ws(char c)
{
  return c==skip_ws_fgetc();
}

bool file::check_next_char(char c)
{
  return c==getc();
}

bool file::check_str_skip_ws(const char *s)
{
  if (!s || !*s) return true;
  
  if (!check_next_char_skip_ws(*s++)) return false;
  
  while (*s) if (!check_next_char(*s++)) return false;
  
  return true;
}

// parses an integer
bool file::parse_int(int &i)
{
  int c;
  
  c=skip_ws_fgetc();
  if (c==EOF) return false;
    
  ungetc(c);
  if (fscanf(f,"%i",&i)!=1) return false;
  
  return true;
}

// parses a unicode character description, as used by the .trans files.
// Valid input is of the form: "U+XXXX" where X is a hex digit.
bool file::parse_uni(unsigned &i)
{
  if (!check_next_char_skip_ws('U') || !check_next_char('+')) return false;
  if (fscanf(f,"%x",&i)!=1) return false;

  return true;
}

// parses a sequence of unicode character descriptions
bool file::parse_uni_str(uni_str &s)
{
  int len=0;
  int alloced=0;
  unsigned *buff=NULL;
  unsigned i;
  char c;
  
  c=skip_ws_fgetc();
  while (c=='U') {
    ungetc('U');
    if (!parse_uni(i)) {
      free(buff);
      return false;
    }
    len++;
    if (len>alloced) {
      alloced+=10;
      buff=(unsigned*)realloc(buff,sizeof(unsigned)*alloced);
    }
    buff[len-1]=i;

    c=skip_ws_fgetc();
  }
  
  s.set(buff,len);
  free(buff);
  
  if (c!=EOF) ungetc(c);
  return true;
}

int file::get_line_no()
{
  return line;
}

bool file::ferror()
{
  return !f || ::ferror(f);
}

bool file::feof()
{
  return !f || ::feof(f);
}

/*
  Accept the format as used by kbdtools:
  
  # denotes comment
  other lines:
  num1 whitespace num2
  
  num1=src
  num2=desc

 This is actually a bit sloppy - it ignores newlines, so it accepts any
 sequence of <int> <uni> items. But this includes well-formet uni translation
 tables, so I think it's ok. :)
*/
void read_table(FILE *f_handle,trans_tbl &t)
{
  int c;
  int from;
  uni_str to;
  static const char err_transtbl[]="Error in translation table at line %d.\n";
  static const char warn_reset[]="Warning: resetting the value of byte %02x at line %d.\n";
  static const char err_out_of_range[]="A source byte in translation table is ouside the 0x00 - 0xFF range at line %d.\n";
  
  file f(f_handle);
  
  while (!f.feof() && !f.ferror()) {
    if (!f.parse_int(from)) {
      if (f.feof()) {
        break;
      }
      fprintf(stderr,err_transtbl,f.get_line_no());
      exit(1);
    }
    
    // check for possible range: %i - %i
    c=f.skip_ws_fgetc();
    
    if (c=='-') {
      // this is a range
      int start=from,end;
      
      if (!f.parse_int(end)) {
        fprintf(stderr,err_transtbl,f.get_line_no());
        exit(1);
      }
      
      //fprintf(stderr,"Range detected: %02x - %02x\n",start,end);
      
      // we detected a range, the only keyword here we know is "idem"
      if (!f.check_str_skip_ws("idem")) {
        fprintf(stderr,"Error in translation table at line %d: 'idem' expected.",f.get_line_no());
        exit(1);
      }

      // check byte values
      if (start<0 || start>=0x100 || end<0 || end>=0x100 || start>end) {
        fprintf(stderr,err_out_of_range,f.get_line_no());
        exit(1);
      }

      // set all bytes in the range to itentity
      for (int i=start;i<=end;i++) {
        if (t.getAt(i)) fprintf(stderr,warn_reset,from,f.get_line_no());
        to.set((unsigned*)&i,1); // one unsigned value
        t.setAt(i,to);
      }
    } else {
      // standard item
      f.ungetc(c);
      
      if (!f.parse_uni_str(to)) {
        fprintf(stderr,err_transtbl,f.get_line_no());
        exit(1);
      }

      // check byte values
      if (from<0 || from>=0x100) {
        fprintf(stderr,err_out_of_range,f.get_line_no());
        exit(1);
      }
      
      if (to.length()==0) {
        fprintf(stderr,err_transtbl,f.get_line_no());
        exit(1);
      }
    
      if (t.getAt(from)) fprintf(stderr,warn_reset,from,f.get_line_no());
    
      t.setAt(from,to);
    }
  }
  
  for (int i=0;i<0x100;i++)
    if (!t.getAt(i)) fprintf(stderr,"Warning: no character assigned to byte %02x.\n",i);
}
